import { getDocument, GlobalWorkerOptions, InvalidPDFException, UnexpectedResponseException, createPromiseCapability, AnnotationEditorUIManager } from 'pdfjs-dist/legacy/build/pdf'
import CryptoJS from 'crypto-js';
import printJS from 'print-js';
import { saveAs } from 'file-saver';
import { PDFLinkService } from "./pdf_link_service"
import { PasswordPrompt } from "./password_prompt.js";
import { PDFFindBar } from "./pdf_find_bar.js";
import { PDFFindController } from "./pdf_find_controller.js";
import { PDFViewer } from './pdf_viewer'
import { PDFThumbnailViewer } from './pdf_thumbnail_viewer'
import { PDFRenderingQueue } from "./pdf_rendering_queue";
import { PDFSidebar } from "./pdf_sidebar.js";
import { GenericExternalServices } from "./genericcom.js";
import {  PDFCursorTools } from "./pdf_cursor_tools.js";
import { EventBus, isValidRotation, isValidScrollMode, shadow, isValidSpreadMode, findIndex, scrollIntoView, DEFAULT_SCALE_VALUE, SidebarView, parseAnnotationFromXml, PromiseExt, DEFAULT_SCALE, UNKNOWN_SCALE, PresentationModeState, watchScroll, ScrollMode, SpreadMode, getVisibleElements, DEFAULT_CACHE_SIZE, AnnotationEditorType } from './ui_utils'
import { AnnotationTypeString, PixelsPerInch } from "../constants";
import { PDFPresentationMode } from "./pdf_presentation_mode.js";
import annotationStore from "./annotation_store"
import { InkSign } from "./ink_sign"
import MessageHandler from "./message_handler"
import JSZip from 'jszip'
import Outline from './Outline'
import PDFPageViewBuffer from './CPDFPageViewBuffer'
import { PDFPageView } from "./pdf_page_view.js"
import { convertFileToBuffer } from './fileHandler'
import { getPdfFilenameFromUrl } from './utils';

GlobalWorkerOptions.workerSrc = './lib/pdf.worker.min.js'
const CMAP_URL = './cmaps/'

function isValidAnnotationEditorMode(mode) {
  return (
    Object.values(AnnotationEditorType).includes(mode) &&
    mode !== AnnotationEditorType.DISABLE
  );
}

class ComPDFKitViewer {
  #license = null
  #doc = null
  #pwd = ''
  #oldPwd = ''
  #buffer = null
  #scrollModePageState = null
  #onVisibilityChange = null
  #scaleTimeoutId = null
  _spreadMode = SpreadMode.NONE
  _scrollMode = ScrollMode.VERTICAL
  _pages = []
  constructor(options) {
    this.viewer = null
    this.pdfLoadingTask = null
    this.printService = null
    this.pdfSidebar = null
    this.outlines = []
    this.pdfLayerViewer = null
    this.pdfDocument = null
    this.downloadComplete = false
    this.isInitialViewSet = false
    this.pdfLinkService = null
    this.externalServices = GenericExternalServices
    this.percent = 0
    this.pageChangedCallback = null
    this.scaleChangedCallback = null
    this.annotationsNumChangedCallback = null
    this.distanceChangedCallback = null
    this._license = ''
    this._token = null
    this._pdfId = null
    this.annotations = null
    this.annotationHistory = []
    this._annotationMode = 0
    this._boundEvents = Object.create(null)
    this._title = document.title
    this.activeTool = ''
    this.color = 'color'
    this.pagesPtr = []
    this.saveAction = null
    this.optionUrl = {
      baseUrl: 'https://test-compdf.kdan.cn',
      // baseUrl: 'https://wms.compdf.com',
      verifyUrl: '/api/license/verify',
      webviewBaseUrl: 'https://test-pdf-pro.kdan.cn:3032',
      // webviewBaseUrl: 'https://web-viewer-backend.compdf.com',
      uploadUrl: '/web-viewer/annotate/openPdfToAnnotate',
      editUrl: '/web-viewer/annotate/edit',
      saveUrl: '/web-viewer/annotate/save',
      flattenUrl: '/web-viewer/annotate/saveAndFlatten',
      uploadAnnotationsUrl: '/web-viewer/annotate/importAnnotateToPdf',
      downloadAnnotation: '/web-viewer/annotate/exportAnnotate',
      compareUrl: '/web-viewer/annotate/pdfCompare'
    }

    this.eventBus = new EventBus();
    this.toolMode = ''
    this.$t = options.$t
    this.fontFileInited = false

    this.annotationStore = annotationStore

    this.linkService = null
    this.findController = null

    this._fontFile = null
    this.presentationModeState = PresentationModeState.UNKNOWN
    this._onBeforeDraw = this._onAfterDraw = null
  }

  async initConfig(options) {
    if (!options.license) return false
    this.#license = options.license
    const data = {
      method: 'POST',
      headers: {
        'Content-Type': 'application/json'
      },
      body: JSON.stringify({
        license: this.#license
      })
    }
    
    await new Promise((resolve, reject) => {
      const worker = new PDFWorker()

      this.worker = worker
      const messageHandler = worker.messageHandler
      this.messageHandler = messageHandler
      messageHandler.on('initDoc', ({ doc }) => {
        console.log('initDoc')
        this.#doc = doc
        resolve(true)
      })
    })
    return fetch(this.optionUrl.baseUrl + this.optionUrl.verifyUrl, data)
      .then((res) => {
        return res.json()
      })
      .then((resp) => {
        if (resp.code === 200) {
          const verified = this._decrypt(resp.data)
          return verified
        } else {
          return false
        }
      })
      .catch((error) => ({ error: true, message: error.message }))
  }

  _decrypt(data) {
    const decrypt = CryptoJS.AES.decrypt(data, CryptoJS.enc.Utf8.parse('T2EKWNOflrd6ICU5'), {
      iv: CryptoJS.enc.Utf8.parse('GM3Faw9kX7CBmojt')
    })
    const decrypted = JSON.parse(decrypt.toString(CryptoJS.enc.Utf8))
    const hostname = window.location.hostname
    const reg = new RegExp(`^.*${decrypted.domain}$`);
    if (reg.test(hostname) || data.app_type === 2) {
      this._token = decrypted.token
      return true
    }
    return false
  }

  initializeViewer({
    container,
    viewer,
    thumbnailView,
    findbarView,
    toggleButton,
  }) {
    this._initializeViewerComponents({
      container,
      viewer,
      thumbnailView,
      findbarView,
      toggleButton,
    })
    this.initBindEvents()
    if (thumbnailView) {
      this.bindEvents();
      this.bindWindowEvents();
    }
  }

  _initializeViewerComponents({
    container,
    viewer,
    thumbnailView,
    findbarView,
    toggleButton,
  }) {
    this.container = container
    this.viewer = viewer
    this.scroll = watchScroll(container, this._scrollUpdate.bind(this))
    const pdfRenderingQueue = new PDFRenderingQueue()
    pdfRenderingQueue.forceRenderingPages = this.forceRenderingPages.bind(this)
    pdfRenderingQueue.onIdle = this._cleanup.bind(this)
    this.pdfRenderingQueue = pdfRenderingQueue

    const pdfLinkService = new PDFLinkService()
    this.pdfLinkService = pdfLinkService

    const findController = new PDFFindController({
      linkService: pdfLinkService,
      eventBus: this.eventBus,
      updateMatchesCountOnProgress: true
    })

    this.findController = findController

    this._resetView()

    pdfLinkService.setViewer(this)

    if (thumbnailView) {
      this.pdfThumbnailViewer = new PDFThumbnailViewer({
        container: thumbnailView,
        renderingQueue: pdfRenderingQueue,
        linkService: pdfLinkService
      });
      pdfRenderingQueue.setThumbnailViewer(this.pdfThumbnailViewer);
    }

    this.findBar = new PDFFindBar({
      bar: document.getElementById("findbar"),
      toggleButton: document.getElementById("viewFind"),
      findField: document.getElementById("findInput"),
      emptyInput: document.getElementById("emptyInput"),
      highlightAllCheckbox: document.getElementById("findHighlightAll"),
      caseSensitiveCheckbox: document.getElementById("findMatchCase"),
      matchDiacriticsCheckbox: document.getElementById("findMatchDiacritics"),
      entireWordCheckbox: document.getElementById("findEntireWord"),
      findMsg: document.getElementById("findMsg"),
      findResultsCount: document.getElementById("findResultsCount"),
      findPreviousButton: document.getElementById("findPrevious"),
      findNextButton: document.getElementById("findNext"),
      findbarView,
      $t: this.$t
    }, this.eventBus);

    this.pdfCursorTools = new PDFCursorTools({
      container,
      eventBus: this.eventBus,
      cursorToolOnLoad: 0,
    });

    if (this.supportsFullscreen) {
      this.pdfPresentationMode = new PDFPresentationMode({
        container,
        documentViewer: this.documentViewer,
        eventBus: this.eventBus,
      });
    }

    this.passwordPrompt = new PasswordPrompt(
      {
        dialog: document.getElementById("passwordDialog"),
        label: document.getElementById("passwordText"),
        input: document.getElementById("password"),
        submitButton: document.getElementById("passwordSubmit"),
        cancelButton: document.getElementById("passwordCancel"),
        $t: this.$t
      }
    );

    if (toggleButton) {
      this.pdfSidebar = new PDFSidebar({
        elements: {
          toggleButton
        },
        documentViewer: this.documentViewer,
        pdfThumbnailViewer: this.pdfThumbnailViewer,
        eventBus: this.eventBus
      });
      this.pdfSidebar.onToggled = this.forceRendering.bind(this);
    }
  }

  async loadDocument(file, options = {}) {
    if (this.pdfLoadingTask) {
      await this.close()
    }
    this.info = options.info || null
    this.pageChangedCallback = options.pageChangedCallback || this.pageChangedCallback
    this.scaleChangedCallback = options.scaleChangedCallback || this.scaleChangedCallback
    this.annotationsNumChangedCallback = options.annotationsNumChangedCallback || this.annotationsNumChangedCallback
    this.distanceChangedCallback = options.distanceChangedCallback || this.distanceChangedCallback
    annotationStore.annotationsAll = null
    annotationStore.selectedName = null
    this.fontFileInited = false
    // 获取pdf信息，返回promise
    this._docName = options.filename || getPdfFilenameFromUrl(file)

    let fileData = file
    
    if (typeof file === 'string') {
      fileData = await this.fetchPdf({
        url: file,
      })
    }
    const buffer = await convertFileToBuffer(fileData)
    await this.messageHandler?.sendWithPromise('LoadFile', {
      buffer,
    })

    this.messageHandler.on('DocumentLoaded', (data) => {
      this.load()
    })

    this.messageHandler.on('PasswordRequest', (data) => {
      const passwordCapability = new PromiseExt()
      this.isEncrypted = true

      const checkPassword = (password) => {
        passwordCapability.resolve(password)
      }

      options.password(checkPassword)

      return passwordCapability.promise
    })

    const loadRes = await this.messageHandler.sendWithPromise('LoadDocumentByStream', {
      doc: this.doc,
      fileId: 0,
      length: buffer.length,
      password: ''
    })

    // notUploadPdf -> 仅预览 不上传到服务器
    // if (options.notUploadPdf) {
    //   const pdfDocument = await loadingTask.promise;
    //   await this.load(pdfDocument);
    //   return this.#pwd
    // }

    // 处理传入的PDF
    return await loadingTask.promise.then(
      async pdfDocument => {
        // 上传pdf
        await pdfDocument.getData().then(async (binaryData) => {
          this.pdfDocument = pdfDocument
          if (!options.notUpdatePwd) {
            this.#oldPwd = this.#pwd
          }
          const formData = new FormData();
          formData.append('file', new Blob([binaryData], { type: 'application/pdf' }), this._docName);
          formData.append('password', this.#pwd);

          let buffer = new Uint8Array(binaryData)
          this.messageHandler.sendWithPromise('LoadFile', {
            buffer
          })

          const loadRes = await this.messageHandler.sendWithPromise('LoadDocumentByStream', {
            doc: this.#doc,
            fileId: 0,
            length: buffer.length,
            password: this.#pwd
          })
          console.log(loadRes)
          if (!loadRes) {
            const pageCount = await this.messageHandler.sendWithPromise('GetPageCount', {
              doc: this.#doc
            })
            this._pageCount = pageCount
          }

          const annotations = await this.getAnnotations()
          this.initAnnotations(annotations)
          // const data = {
          //   method: 'POST',
          //   headers: {
          //     'Authorization': this._token
          //   },
          //   body: formData
          // }
          // await fetch(this.optionUrl.webviewBaseUrl + this.optionUrl.uploadUrl, data)
          //   .then((res) => {
          //     return res.json()
          //   })
          //   .then((resp) => {
          //     if (resp.code === "200") {
          //       const data = resp.data
          //       this._pdfId = data.pdfId
          //       const annotations = data.annotateJson && JSON.parse(data.annotateJson)
          //       this.initAnnotations(annotations)
          //     } else if (resp.code === "319") {

          //     }
          //   })
          //   .catch((error) => {
          //     console.log(error)
          //     return { error: true, message: error.message }
          //   });
        })
        await this.load(pdfDocument);
        return { pwd: !!this.#pwd };
      },
      reason => {
        if (loadingTask !== this.pdfLoadingTask) {
          return undefined; // Ignore errors for previously opened PDF files.
        }

        let key = "loading_error";
        if (reason instanceof InvalidPDFException) {
          key = "invalid_file_error";
          console.warn('Invalid or corrupted PDF file.')
          return new Promise(function (resolve, reject) {
            reject('invalid_file_error')
          });
        } else if (reason instanceof MissingPDFException) {
          key = "missing_file_error";
          console.warn('Missing PDF file.')
        } else if (reason instanceof UnexpectedResponseException) {
          key = "unexpected_response_error";
          console.warn('Unexpected server response.')
        } else if (reason) {
          if (reason.message === 'No password given') {
            key = "no_password_given";
            console.warn('No password given.')
            return new Promise(function (resolve, reject) {
              reject('no_password_given')
            });
          } else if (reason.message === 'Incorrect Password') {
            return new Promise(function (resolve, reject) {
              reject('incorrect_password')
            });
          }
        }
        throw new Error(key)
      }
    )
  }

  fetchPdf(requestData, withCredentials = false) {
    const { url, customHeaders } = requestData

    const headers = this.createHeaders(customHeaders)

    const abortController = new AbortController()

    const options = this.createFetchOptions(headers, withCredentials, abortController)
    const promiseCapability = new PromiseExt()

    fetch(url, options)
      .then((response) => {
        if (!response.ok) {
          return promiseCapability.reject({
            message: "Received http error code " + response.status
          })
        }
        const contentDisposition = response.headers.get("Content-Disposition")
        if (contentDisposition) {
          if (contentDisposition.includes("%")) {
            try {
              this._filename = decodeURIComponent(contentDisposition);
            } catch { }
          } else {
            this._filename = contentDisposition
          }
        } else {
          this._filename = getPdfFilenameFromUrl(url)
        }

        const type = response.headers.get("Content-Type")
        if (type?.startsWith('application/json')) {
          const errorObj = res.json();
          promiseCapability.reject({
            message: errorObj
          })
        } else {
          promiseCapability.resolve(response.blob())
        }
      })
      .catch((error) => {
        console.log(error)
        promiseCapability.reject({
          message: "Network or Cross Domain Access Error."
        })
      })

    return promiseCapability.promise
  }

  createHeaders(httpHeaders) {
    const headers = new Headers();
    for (const property in httpHeaders) {
      const value = httpHeaders[property];
      if (value === undefined) {
        continue;
      }
      headers.append(property, value);
    }
    return headers;
  }

  createFetchOptions(headers, withCredentials, abortController) {
    return {
      method: "GET",
      headers,
      signal: abortController.signal,
      mode: "cors",
      credentials: withCredentials ? "include" : "same-origin",
      redirect: "follow",
    };
  }

  // 渲染PDF
  async load(pdfDocument) {
    this.setDocument(pdfDocument, this.annotations, this.pagesPtr);

    const pagesCount = await this.messageHandler.sendWithPromise('GetPageCount', {
      doc: this.doc
    })

    for (let pageIndex = 0; pageIndex < pagesCount; pageIndex++) {
      const page = new Page({
        doc: this.doc,
        viewer: this.viewer,
        pagesCount,
        pageIndex,
        zoom: this.zoom,
        rotation: this.rotation,
        messageHandler: this.messageHandler
      })

      await page.init()
      this._pages.push(page)
    }

    this.downloadComplete = true;
    this.eventBus.dispatch("documentloaded", { source: this });
    if (this.pdfThumbnailViewer) {
      const pdfThumbnailViewer = this.pdfThumbnailViewer;
      pdfThumbnailViewer.setDocument(pdfDocument);
    }

    const doc = this.#doc
    const messageHandler = this.messageHandler
    const outlines = await messageHandler.sendWithPromise('GetOutlines', {
      doc
    })

    function generateOutline(outline, index, parent) {
      const outlineItem = new Outline({
        outline,
        index,
        doc,
        messageHandler,
        parent,
        document: this
      })
      outline.children.forEach(function (outline, index) {
        outlineItem.children.push(generateOutline(outline, index, outlineItem))
      })

      return outlineItem
    }

    outlines.forEach((outline, index) => {
      const outlineItem = generateOutline(outline, index, null)
      this.outlines.push(outlineItem)
    })

    let scrollMode = -1
    let spreadMode = -1
    this.setInitialView({
      scrollMode,
      spreadMode,
    });

    this.eventBus.dispatch("documentinit", { source: this });
  }

  getOutlines() {
    if (this.outlines) {
      return this.outlines
    }
  }

  get pageCount() {
    return this.#doc ? this._pages.length : 0;
  }

  /**
   * @type {number}
   */
  get currentPageNumber() {
    return this._currentPageNumber
  }

  /**
   * @param {number} val - The page number.
   */
  set currentPageNumber(val) {
    if (!Number.isInteger(val)) {
      throw new Error("Invalid page number.");
    }

    if (!this.#setCurrentPageNumber(val, /* resetCurrentPageView = */ true)) {
      console.error(`currentPageNumber: "${val}" is not a valid page.`);
    }
  }

  /**
   * @returns {boolean} Whether the pageNumber is valid (within bounds).
   * @private
   */
  #setCurrentPageNumber(val, resetCurrentPageView = false) {
    if (this._currentPageNumber === val) {
      if (resetCurrentPageView) {
        this.#resetCurrentPageView();
      }
      return true;
    }

    if (!(0 < val && val <= this.pageCount)) {
      return false;
    }
    const previous = this._currentPageNumber;
    this._currentPageNumber = val;

    this.eventBus.dispatch("pagechanging", {
      source: this,
      pageNumber: val,
      previous,
    });

    if (resetCurrentPageView) {
      this.#resetCurrentPageView();
    }
    return true;
  }


  get annotationsCount() {
    return this.pdfViewer.annotationStorage ? this.pdfViewer.annotationStorage.size : 0;
  }

  get scale() {
    return this.currentScale;
  }

  /**
   * @type {number}
   */
  get currentScale() {
    return this._currentScale !== UNKNOWN_SCALE
      ? this._currentScale
      : DEFAULT_SCALE;
  }

  /**
   * @param {number} val - Scale of the pages in percents.
   */
  set currentScale(val) {
    if (isNaN(val)) {
      throw new Error("Invalid numeric scale.");
    }
    if (!this.pdfDocument) {
      return;
    }
    this._setScale(val, { noScroll: false });
  }

  /**
   * @type {string}
   */
  get currentScaleValue() {
    return this._currentScaleValue;
  }

  /**
   * @param val - The scale of the pages (in percent or predefined value).
   */
  set currentScaleValue(val) {
    if (!this.pdfDocument) {
      return;
    }
    this._setScale(val, { noScroll: false });
  }

  /**
   * @type {number}
   */
  get pagesRotation() {
    return this._pagesRotation;
  }

  /**
   * @param {number} rotation - The rotation of the pages (0, 90, 180, 270).
   */
  set pagesRotation(rotation) {
    if (!isValidRotation(rotation)) {
      throw new Error("Invalid pages rotation angle.");
    }
    if (!this.pdfDocument) {
      return;
    }
    // Normalize the rotation, by clamping it to the [0, 360) range.
    rotation %= 360;
    if (rotation < 0) {
      rotation += 360;
    }
    if (this._pagesRotation === rotation) {
      return; // The rotation didn't change.
    }
    this._pagesRotation = rotation;

    const pageNumber = this._currentPageNumber;

    this.refresh(true, { rotation });

    // Prevent errors in case the rotation changes *before* the scale has been
    // set to a non-default value.
    if (this._currentScaleValue) {
      this._setScale(this._currentScaleValue, { noScroll: true });
    }

    this.eventBus.dispatch("rotationchanging", {
      source: this,
      pagesRotation: rotation,
      pageNumber,
    });

    if (this.defaultRenderingQueue) {
      this.update();
    }
  }

  #layerProperties() {
    const self = this;
    return {
      get findController() {
        return self.findController;
      },
      get annotationEditorUIManager() {
        return self._annotationEditorUIManager;
      },
      get annotationStorage() {
        return self.pdfDocument?.annotationStorage;
      }
    };
  }

  #onePageRenderedOrForceFetch() {
    if (
      document.visibilityState === "hidden" ||
      !this.container.offsetParent ||
      this._getVisiblePages().views.length === 0
    ) {
      return Promise.resolve();
    }

    const visibilityChangePromise = new Promise(resolve => {
      this.#onVisibilityChange = () => {
        if (document.visibilityState !== "hidden") {
          return;
        }
        resolve();

        document.removeEventListener(
          "visibilitychange",
          this.#onVisibilityChange
        );
        this.#onVisibilityChange = null;
      };
      document.addEventListener("visibilitychange", this.#onVisibilityChange);
    });

    return Promise.race([
      this._onePageRenderedCapability.promise,
      visibilityChangePromise,
    ]);
  }

  #resetCurrentPageView() {
    const pageView = this._pages[this._currentPageNumber - 1]

    if (this.isInPresentationMode) {
      // Fixes the case when PDF has different page sizes.
      this._setScale(this._currentScaleValue, { noScroll: true })
    }
    this.#scrollIntoView(pageView);
  }

  #scrollIntoView(pageView, pageSpot = null) {
    const { div, id } = pageView;

    // Ensure that `this._currentPageNumber` is correct, when `#scrollIntoView`
    // is called directly (and not from `#resetCurrentPageView`).
    if (this._currentPageNumber !== id) {
      this.#setCurrentPageNumber(id);
    }
    if (this._scrollMode === ScrollMode.PAGE) {
      this.#ensurePageViewVisible();
      // Ensure that rendering always occurs, to avoid showing a blank page,
      // even if the current position doesn't change when the page is scrolled.
      this.update();
    }

    if (!pageSpot && !this.isInPresentationMode) {
      const left = div.offsetLeft + div.clientLeft,
        right = left + div.clientWidth;
      const { scrollLeft, clientWidth } = this.container;
      if (
        this._scrollMode === ScrollMode.HORIZONTAL ||
        left < scrollLeft ||
        right > scrollLeft + clientWidth
      ) {
        pageSpot = { left: 0, top: 0 };
      }
    }
    scrollIntoView(div, pageSpot);

    // Ensure that the correct *initial* document position is set, when any
    // OpenParameters are used, for documents with non-default Scroll/Spread
    // modes (fixes issue 15695). This is necessary since the scroll-handler
    // invokes the `update`-method asynchronously, and `this._location` could
    // thus be wrong when the initial zooming occurs in the default viewer.
    if (!this._currentScaleValue && this._location) {
      this._location = null;
    }
  }

  #ensurePageViewVisible() {
    if (this._scrollMode !== ScrollMode.PAGE) {
      throw new Error("#ensurePageViewVisible: Invalid scrollMode value.");
    }
    const pageNumber = this._currentPageNumber,
      state = this.#scrollModePageState,
      viewer = this.viewer;

    // Temporarily remove all the pages from the DOM...
    viewer.textContent = "";
    // ... and clear out the active ones.
    state.pages.length = 0;

    if (this._spreadMode === SpreadMode.NONE && !this.isInPresentationMode) {
      // Finally, append the new page to the viewer.
      const pageView = this._pages[pageNumber - 1];
      viewer.append(pageView.div);

      state.pages.push(pageView);
    } else {
      const pageIndexSet = new Set(),
        parity = this._spreadMode - 1;

      // Determine the pageIndices in the new spread.
      if (parity === -1) {
        // PresentationMode is active, with `SpreadMode.NONE` set.
        pageIndexSet.add(pageNumber - 1);
      } else if (pageNumber % 2 !== parity) {
        // Left-hand side page.
        pageIndexSet.add(pageNumber - 1);
        pageIndexSet.add(pageNumber);
      } else {
        // Right-hand side page.
        pageIndexSet.add(pageNumber - 2);
        pageIndexSet.add(pageNumber - 1);
      }

      // Finally, append the new pages to the viewer and apply the spreadMode.
      const spread = document.createElement("div");
      spread.className = "spread";

      if (this.isInPresentationMode) {
        const dummyPage = document.createElement("div");
        dummyPage.className = "dummyPage";
        spread.append(dummyPage);
      }

      for (const i of pageIndexSet) {
        const pageView = this._pages[i];
        if (!pageView) {
          continue;
        }
        spread.append(pageView.div);

        state.pages.push(pageView);
      }
      viewer.append(spread);
    }

    state.scrollDown = pageNumber >= state.previousPageNumber;
    state.previousPageNumber = pageNumber;
  }

  /**
   * Prevent unnecessary re-rendering of all pages when the scale changes
   * only because of limited numerical precision.
   */
  #isSameScale(newScale) {
    return (
      newScale === this._currentScale ||
      Math.abs(newScale - this._currentScale) < 1e-15
    );
  }

  /**
   * @private
   */
  get _pageWidthScaleFactor() {
    if (
      this._spreadMode !== SpreadMode.NONE &&
      this._scrollMode !== ScrollMode.HORIZONTAL
    ) {
      return 2;
    }
    return 1;
  }

  _setScale(value, options) {
    let scale = parseFloat(value);

    console.log(scale)
    if (scale > 0) {
      options.preset = false;
      this._setScaleUpdatePages(scale, value, options);
    } else {
      const currentPage = this._pages[this._currentPageNumber - 1];
      if (!currentPage) {
        return;
      }
      let hPadding = SCROLLBAR_PADDING,
        vPadding = VERTICAL_PADDING;

      if (this.isInPresentationMode) {
        // Pages have a 2px (transparent) border in PresentationMode, see
        // the `web/pdf_viewer.css` file.
        hPadding = vPadding = 4; // 2 * 2px
        if (this._spreadMode !== SpreadMode.NONE) {
          // Account for two pages being visible in PresentationMode, thus
          // "doubling" the total border width.
          hPadding *= 2;
        }
      } else if (this._scrollMode === ScrollMode.HORIZONTAL) {
        [hPadding, vPadding] = [vPadding, hPadding]; // Swap the padding values.
      }
      const pageWidthScale =
        (((this.container.clientWidth - hPadding) / currentPage.width) *
          currentPage.scale) /
        this._pageWidthScaleFactor;
      const pageHeightScale =
        ((this.container.clientHeight - vPadding) / currentPage.height) *
        currentPage.scale;
      switch (value) {
        case "page-actual":
          scale = 1;
          break;
        case "page-width":
          scale = pageWidthScale;
          break;
        case "page-height":
          scale = pageHeightScale;
          break;
        case "page-fit":
          scale = Math.min(pageWidthScale, pageHeightScale);
          break;
        case "auto":
          // For pages in landscape mode, fit the page height to the viewer
          // *unless* the page would thus become too wide to fit horizontally.
          const horizontalScale = isPortraitOrientation(currentPage)
            ? pageWidthScale
            : Math.min(pageHeightScale, pageWidthScale);
          scale = Math.min(MAX_AUTO_SCALE, horizontalScale);
          break;
        default:
          console.error(`_setScale: "${value}" is an unknown zoom value.`);
          return;
      }
      options.preset = true;
      this._setScaleUpdatePages(scale, value, options);
    }
    this.MarkupAnnotation.scale = this.currentScale
  }

  _setScaleUpdatePages(
    newScale,
    newValue,
    { noScroll = false, preset = false, drawingDelay = -1 }
  ) {
    this._currentScaleValue = newValue.toString();

    if (this.#isSameScale(newScale)) {
      if (preset) {
        this.eventBus.dispatch("scalechanging", {
          source: this,
          scale: newScale,
          presetValue: newValue,
        });
      }
      return;
    }

    this.viewer.style.setProperty(
      "--scale-factor",
      newScale * PixelsPerInch.PDF_TO_CSS_UNITS
    );

    const postponeDrawing = drawingDelay >= 0 && drawingDelay < 1000;
    this.refresh(true, {
      scale: newScale,
      drawingDelay: postponeDrawing ? drawingDelay : -1,
    });

    if (postponeDrawing) {
      this.#scaleTimeoutId = setTimeout(() => {
        this.#scaleTimeoutId = null;
        this.refresh();
      }, drawingDelay);
    }

    this._currentScale = newScale;

    if (!noScroll) {
      let page = this._currentPageNumber,
        dest;
      if (
        this._location &&
        !(this.isInPresentationMode || this.isChangingPresentationMode)
      ) {
        page = this._location.pageNumber;
        dest = [
          null,
          { name: "XYZ" },
          this._location.left,
          this._location.top,
          null,
        ];
      }
      this.scrollPageIntoView({
        pageNumber: page,
        destArray: dest,
        allowNegativeOffset: true,
      });
    }

    this.eventBus.dispatch("scalechanging", {
      source: this,
      scale: newScale,
      presetValue: preset ? newValue : undefined,
    });

    if (this.defaultRenderingQueue) {
      this.update();
    }
  }

  /**
   * Scrolls page into view.
   * @param {ScrollPageIntoViewParameters} params
   */
  scrollPageIntoView({
    pageNumber,
    destArray = null,
    allowNegativeOffset = false,
    ignoreDestinationZoom = false,
  }) {
    if (!this.pdfDocument) {
      return;
    }
    const pageView =
      Number.isInteger(pageNumber) && this._pages[pageNumber - 1];
    if (!pageView) {
      console.error(
        `scrollPageIntoView: "${pageNumber}" is not a valid pageNumber parameter.`
      );
      return;
    }

    if (this.isInPresentationMode || !destArray) {
      this._setCurrentPageNumber(pageNumber, /* resetCurrentPageView = */ true);
      return;
    }
    let x = 0,
      y = 0;
    let width = 0,
      height = 0,
      widthScale,
      heightScale;
    const changeOrientation = pageView.rotation % 180 !== 0;
    const pageWidth =
      (changeOrientation ? pageView.height : pageView.width) /
      pageView.scale /
      PixelsPerInch.PDF_TO_CSS_UNITS;
    const pageHeight =
      (changeOrientation ? pageView.width : pageView.height) /
      pageView.scale /
      PixelsPerInch.PDF_TO_CSS_UNITS;
    let scale = 0;
    switch (destArray[1].name) {
      case "XYZ":
        x = destArray[2];
        y = destArray[3];
        scale = destArray[4];
        // If x and/or y coordinates are not supplied, default to
        // _top_ left of the page (not the obvious bottom left,
        // since aligning the bottom of the intended page with the
        // top of the window is rarely helpful).
        x = x !== null ? x : 0;
        y = y !== null ? y : pageHeight;
        break;
      case "Fit":
      case "FitB":
        scale = "page-fit";
        break;
      case "FitH":
      case "FitBH":
        y = destArray[2];
        scale = "page-width";
        // According to the PDF spec, section 12.3.2.2, a `null` value in the
        // parameter should maintain the position relative to the new page.
        if (y === null && this._location) {
          x = this._location.left;
          y = this._location.top;
        } else if (typeof y !== "number" || y < 0) {
          // The "top" value isn't optional, according to the spec, however some
          // bad PDF generators will pretend that it is (fixes bug 1663390).
          y = pageHeight;
        }
        break;
      case "FitV":
      case "FitBV":
        x = destArray[2];
        width = pageWidth;
        height = pageHeight;
        scale = "page-height";
        break;
      case "FitR":
        x = destArray[2];
        y = destArray[3];
        width = destArray[4] - x;
        height = destArray[5] - y;
        const hPadding = SCROLLBAR_PADDING;
        const vPadding = VERTICAL_PADDING;

        widthScale =
          (this.container.clientWidth - hPadding) /
          width /
          PixelsPerInch.PDF_TO_CSS_UNITS;
        heightScale =
          (this.container.clientHeight - vPadding) /
          height /
          PixelsPerInch.PDF_TO_CSS_UNITS;
        scale = Math.min(Math.abs(widthScale), Math.abs(heightScale));
        break;
      default:
        console.error(
          `scrollPageIntoView: "${destArray[1].name}" is not a valid destination type.`
        );
        return;
    }

    if (!ignoreDestinationZoom) {
      if (scale && scale !== this._currentScale) {
        this.currentScaleValue = scale;
      } else if (this._currentScale === UNKNOWN_SCALE) {
        this.currentScaleValue = DEFAULT_SCALE_VALUE;
      }
    }

    if (scale === "page-fit" && !destArray[4]) {
      this.#scrollIntoView(pageView);
      return;
    }

    const boundingRect = [
      pageView.viewport.convertToViewportPoint(x, y),
      pageView.viewport.convertToViewportPoint(x + width, y + height),
    ];
    let left = Math.min(boundingRect[0][0], boundingRect[1][0]);
    let top = Math.min(boundingRect[0][1], boundingRect[1][1]);

    if (!allowNegativeOffset) {
      // Some bad PDF generators will create destinations with e.g. top values
      // that exceeds the page height. Ensure that offsets are not negative,
      // to prevent a previous page from becoming visible (fixes bug 874482).
      left = Math.max(left, 0);
      top = Math.max(top, 0);
    }
    this.#scrollIntoView(pageView, /* pageSpot = */ { left, top });
  }

  focus() {
    this.container.focus();
  }

  get isInPresentationMode() {
    return this.presentationModeState === PresentationModeState.FULLSCREEN;
  }

  _resetView() {
    this._pages = []
    this._currentPageNumber = 1
    this._currentScale = UNKNOWN_SCALE;
    this._currentScaleValue = null;
    this.#buffer = new PDFPageViewBuffer(DEFAULT_CACHE_SIZE);
    this._pagesRotation = 0;
    this._firstPageCapability = createPromiseCapability();
    this._onePageRenderedCapability = createPromiseCapability();
    this._pagesCapability = createPromiseCapability();
    this.scrollMode = ScrollMode.VERTICAL;
    this._previousScrollMode = ScrollMode.UNKNOWN;
    this._spreadMode = SpreadMode.NONE;

    this.#scrollModePageState = {
      previousPageNumber: 1,
      scrollDown: true,
      pages: [],
    };

    if (this._onBeforeDraw) {
      this.eventBus._off("pagerender", this._onBeforeDraw);
      this._onBeforeDraw = null;
    }
    if (this._onAfterDraw) {
      this.eventBus._off("pagerendered", this._onAfterDraw);
      this._onAfterDraw = null;
    }
    if (this.#onVisibilityChange) {
      document.removeEventListener(
        "visibilitychange",
        this.#onVisibilityChange
      );
      this.#onVisibilityChange = null;
    }
    // Remove the pages from the DOM...
    this.viewer.textContent = "";
    // ... and reset the Scroll mode CSS class(es) afterwards.
    this._updateScrollMode();
  }

  getPageView(index) {
    return this._pages[index]
  }

  setTool({
    tool,
    color
  }) {
    this.activeTool = tool
    this.color = color
    this.eventBus.dispatch("toolChanged", { tool, color });
  }

  initAnnotations(annotations, render = false) {
    for (let index = 0; index < annotations.length; index++) {
      this.handleAnnotations(annotations[index], true)
      render && this.pdfViewer.renderAnnotation(annotations[index])
    }
    console.log(this.annotations)
    this.eventBus.dispatch('annotationChanged', { annotations: this.annotations })
  }

  async handleAnnotations(annotation, init = false) {
    this.annotations = this.annotations || {}
    if (init) {
      this.initAddAnnotations(annotation)
    } else {
      if (Number(annotation.pageIndex) + 1 > this.pageCount) return
      this.pushAnnotations(annotation)
      const pagePtr = this.pagesPtr[annotation.pageIndex].pagePtr
      const annotPtr = await this.messageHandler.sendWithPromise('CreateAnnotation', {
        doc: this.#doc,
        pagePtr,
        annotation
      })
      annotation.pagePtr = pagePtr
      annotation.annotPtr = annotPtr
    }
  }

  initAddAnnotations(annotation) {
    const pageIndex = annotation.pageIndex
    const annotations = this.annotations
    if (!annotations[pageIndex]) {
      annotations[pageIndex] = []
    }

    annotations[pageIndex].push(annotation)
    if (!annotationStore.annotationsAll) {
      annotationStore.annotationsAll = annotations
    }
  }

  pushAnnotations(annotation) {
    const pageIndex = annotation.pageIndex
    const annotations = this.annotations
    if (!annotations[pageIndex]) {
      annotations[pageIndex] = []
      this.pdfViewer._pages[pageIndex].annotations = annotations[pageIndex]
    }
    if (!annotationStore.annotationsAll) {
      annotationStore.annotationsAll = annotations
    }
    annotations[pageIndex].push(annotation)
  }

  addAnnotations(annotation) {
    this.handleAnnotationChange({
      type: 'add',
      annotation: [annotation]
    })
  }

  delAnnotations() {

  }

  addEvent(eventName, fn) {
    this.eventBus._on(eventName, fn);
  }

  async search(value) {
    const searchResults = await this.messageHandler.sendWithPromise('Search', {
      pagesPtr: this.pagesPtr,
      value
    })

    this.searchResults = searchResults

    return searchResults
  }

  async getAnnotations() {
    this.pagesPtr = []
    const annotations = []

    for (let pageIndex = 0; pageIndex < this.pageCount; pageIndex++) {
      const pageSize = await this.messageHandler.sendWithPromise("GetPageSize", {
        doc: this.#doc,
        pageIndex
      })
      const textPtr = await this.messageHandler.sendWithPromise('InitText', {
        pagePtr: pageSize.pagePtr
      })

      this.pagesPtr.push({
        doc: this.#doc,
        pagePtr: pageSize.pagePtr,
        textPtr,
      })
      const data = await this.messageHandler.sendWithPromise('InitAnnot', {
        pageIndex,
        pagePtr: pageSize.pagePtr
      })
      for (let annotCountNum = 0; annotCountNum < data.annotCount; annotCountNum++) {
        const annotation = data.annotations[annotCountNum]
        const typeInt = annotation.type
        const type = AnnotationTypeString[typeInt]
        annotation.pageIndex = pageIndex
        annotation.type = type
        let attr = null
        if (type === 'widget') {
          attr = await this.messageHandler.sendWithPromise('GetWidgetAnnotation', {
            doc: this.#doc,
            annotPtr: annotation.annotPtr,
            pagePtr: annotation.pagePtr,
            typeInt
          })
        } else {
          attr = await this.messageHandler.sendWithPromise('GetAnnotation', {
            doc: this.#doc,
            annotPtr: annotation.annotPtr,
            pagePtr: annotation.pagePtr,
            typeInt,
            scale: this.scale
          })
        }
        Object.assign(annotation, attr)
        annotations.push(annotation)
      }
    }

    this.pdfViewer.pagesPtr = this.pagesPtr

    return annotations
  }

  async reRenderAnnotations() {
    this.annotations = null
    const annotations = []
    const pagesPtr = this.pagesPtr

    for (let pageIndex = 0; pageIndex < pagesPtr.length; pageIndex++) {
      const pagePtr = pagesPtr[pageIndex]
      const data = await this.messageHandler.sendWithPromise('InitAnnot', {
        pageIndex,
        pagePtr: pagePtr.pagePtr
      })
      for (let annotCountNum = 0; annotCountNum < data.annotCount; annotCountNum++) {
        const annotation = data.annotations[annotCountNum]
        const typeInt = annotation.type
        const type = AnnotationTypeString[typeInt]
        annotation.pageIndex = pageIndex
        annotation.type = type
        let attr = null
        if (type === 'widget') {
          attr = await this.messageHandler.sendWithPromise('GetWidgetAnnotation', {
            doc: this.#doc,
            annotPtr: annotation.annotPtr,
            pagePtr: annotation.pagePtr,
            typeInt
          })
        } else {
          attr = await this.messageHandler.sendWithPromise('GetAnnotation', {
            doc: this.#doc,
            annotPtr: annotation.annotPtr,
            pagePtr: annotation.pagePtr,
            typeInt,
            scale: this.scale
          })
        }
        Object.assign(annotation, attr)
        annotations.push(annotation)
      }
    }

    this.initAnnotations(annotations, true)
  }

  get supportsFullscreen() {
    const fullscreenEnabled = document.webkitFullscreenEnabled || document.mozFullscreenEnabled || document.fullscreenEnabled
    return shadow(this, "supportsFullscreen", fullscreenEnabled);
  }

  removeAllAnnotations() {
    if (this.pdfViewer.annotationStorage) {
      this.pdfViewer.annotationStorage.cleanup()
    }
  }

  async importAnnotations(data) {
    const source = await new Promise((resolve) => {
      const reader = new FileReader();
      reader.onload = (event) => {
        const contents = event.target.result
        resolve(contents)
      };

      if (this.webviewerServer) {
        reader.readAsText(data);
      } else {
        reader.readAsArrayBuffer(data);
      }
    });
    if (this.webviewerServer) {
      const annotationsXml = parseAnnotationFromXml(source)
      if (!annotationsXml.length) return
      this.handleAnnotationChange({
        type: 'add',
        annotation: annotationsXml
      })
    } else {
      console.log(source)
      const xfdfBuffer = new Uint8Array(source)
      console.log(xfdfBuffer)
      await this.messageHandler.sendWithPromise('XFDFImportAnnotations', {
        doc: this.#doc,
        xfdfBuffer
      })
    }

    this.reRenderAnnotations()
  }

  setPassword(value) {
    if (this.webviewerServer) {
      const options = {
        method: 'POST',
        headers: {
          "Content-Type": "application/json",
          'Authorization': this._token
        },
        body: JSON.stringify({
          pdfId: this._pdfId,
          annotateHandles: [{
            operate: 'set-password',
            userPassword: value
          }]
        }),
      }

      return fetch(this.optionUrl.webviewBaseUrl + this.optionUrl.editUrl, options)
        .then((res) => {
          return res.json()
        })
        .then((data) => {
          if (data.code === "200") {
            return true
          } else {
            return false
          }
        })
        .catch((error) => {
          console.log(error)
          return false
        });
    } else {
      // 设置密码标志
      this.saveAction = 'set'
      this.#pwd = value
      return true
    }

  }

  removePassword() {
    if (this.webviewerServer) {
      const options = {
        method: 'POST',
        headers: {
          "Content-Type": "application/json",
          'Authorization': this._token
        },
        body: JSON.stringify({
          pdfId: this._pdfId,
          annotateHandles: [{
            operate: 'remove-password'
          }]
        }),
      }

      return fetch(this.optionUrl.webviewBaseUrl + this.optionUrl.editUrl, options)
        .then((res) => {
          return res.json()
        })
        .then((data) => {
          if (data.code === "200") {
            return true
          } else {
            return false
          }
        })
        .catch((error) => {
          console.log(error)
          return false
        });
    } else {
      if (this.#pwd) {
        // 移除密码标志
        this.saveAction = 'remove'
        this.#pwd = ''
      }
      return true
    }
  }

  getOptionUrl() {
    return this.optionUrl
  }

  toggleSidebar() {
    this.pdfSidebar.toggle()
  }

  /**
   * @param {PDFDocumentProxy} pdfDocument
   */
  setDocument(pdfDocument, annotations, pagesPtr) {
    if (this.pdfDocument) {
      this.eventBus.dispatch("pagesdestroy", { source: this });

      this._cancelRendering();
      this._resetView();

      this.findController?.setDocument(null);

      if (this._annotationEditorUIManager) {
        this._annotationEditorUIManager.destroy();
        this._annotationEditorUIManager = null;
      }

      this.pagesPtr = pagesPtr

      if (this.MarkupAnnotation) {
        this.MarkupAnnotation.destroy()
        this.MarkupAnnotation = null
      }
    }

    this.pdfDocument = pdfDocument;
    if (!pdfDocument) {
      return;
    }
    const pagesCount = pdfDocument.numPages;
    const firstPagePromise = pdfDocument.getPage(1);
    const permissionsPromise = Promise.resolve();

    this._pagesCapability.promise.then(
      () => {
        this.eventBus.dispatch("pagesloaded", { source: this, pagesCount });
      },
      () => {
        /* Prevent "Uncaught (in promise)"-messages in the console. */
      }
    );

    this._onBeforeDraw = evt => {
      const pageView = this._pages[evt.pageNumber - 1];
      if (!pageView) {
        return;
      }
      // Add the page to the buffer at the start of drawing. That way it can be
      // evicted from the buffer and destroyed even if we pause its rendering.
      this.#buffer.push(pageView);
    };
    this.eventBus._on("pagerender", this._onBeforeDraw);

    this._onAfterDraw = evt => {
      if (evt.cssTransform || this._onePageRenderedCapability.settled) {
        return;
      }
      this._onePageRenderedCapability.resolve({ timestamp: evt.timestamp });

      this.eventBus._off("pagerendered", this._onAfterDraw);
      this._onAfterDraw = null;

      if (this.#onVisibilityChange) {
        document.removeEventListener(
          "visibilitychange",
          this.#onVisibilityChange
        );
        this.#onVisibilityChange = null;
      }
    };
    this.eventBus._on("pagerendered", this._onAfterDraw);

    // Fetch a single page so we can get a viewport that will be the default
    // viewport for all pages
    Promise.all([firstPagePromise, permissionsPromise])
      .then(([firstPdfPage, permissions]) => {
        if (pdfDocument !== this.pdfDocument) {
          return; // The document was closed while the first page resolved.
        }
        this._firstPageCapability.resolve(firstPdfPage);
        console.log(permissions)
        const annotationEditorMode = 0,
          textLayerMode = 1

        if (annotationEditorMode !== AnnotationEditorType.DISABLE) {
          const mode = annotationEditorMode;

          if (isValidAnnotationEditorMode(mode)) {
            this._annotationEditorUIManager = new AnnotationEditorUIManager(
              this.container,
              this.eventBus,
              pdfDocument?.annotationStorage
            );
          }
        }
        const layerProperties = this.#layerProperties.bind(this);

        const viewerElement =
          this._scrollMode === ScrollMode.PAGE ? null : this.viewer;
        const scale = this.currentScale;
        const viewport = firstPdfPage.getViewport({
          scale: scale * PixelsPerInch.PDF_TO_CSS_UNITS,
        });
        // Ensure that the various layers always get the correct initial size,
        // see issue 15795.
        this.viewer.style.setProperty("--scale-factor", viewport.scale);

        for (let pageNum = 1; pageNum <= pagesCount; ++pageNum) {
          const pageView = new PDFPageView({
            container: viewerElement,
            eventBus: this.eventBus,
            id: pageNum,
            pageIndex: pageNum - 1,
            scale,
            messageHandler: this.messageHandler,
            pagesPtr: this.pagesPtr,
            annotationStore: this.annotationStore,
            annotations: annotations && annotations[pageNum - 1] || null,
            annotationsAll: annotations,
            pagesPtr: this.pagesPtr,
            defaultViewport: viewport.clone(),
            renderingQueue: this.renderingQueue,
            textLayerMode,
            imageResourcesPath: this.imageResourcesPath,
            renderer:
              typeof PDFJSDev === "undefined" ||
                PDFJSDev.test("!PRODUCTION || GENERIC")
                ? this.renderer
                : null,
            useOnlyCssZoom: this.useOnlyCssZoom,
            isOffscreenCanvasSupported: this.isOffscreenCanvasSupported,
            maxCanvasPixels: this.maxCanvasPixels,
            layerProperties,
            l10n: this.l10n,
            $t: this.$t,
            doc: this.doc,
            messageHandler: this.messageHandler
          });
          this._pages.push(pageView);
        }
        // Set the first `pdfPage` immediately, since it's already loaded,
        // rather than having to repeat the `PDFDocumentProxy.getPage` call in
        // the `this.#ensurePdfPageLoaded` method before rendering can start.
        const firstPageView = this._pages[0];
        if (firstPageView) {
          firstPageView.setPdfPage(firstPdfPage);
        }

        this.MarkupAnnotation = new MarkupAnnotation({
          eventBus: this.eventBus,
          viewport,
          pages: this._pages,
          scale: this.currentScale
        })

        if (this._scrollMode === ScrollMode.PAGE) {
          // Ensure that the current page becomes visible on document load.
          this.#ensurePageViewVisible();
        } else if (this._spreadMode !== SpreadMode.NONE) {
          this._updateSpreadMode();
        }

        // Fetch all the pages since the viewport is needed before printing
        // starts to create the correct size canvas. Wait until one page is
        // rendered so we don't tie up too many resources early on.
        this.#onePageRenderedOrForceFetch().then(async () => {
          this.findController?.setDocument(pdfDocument); // Enable searching.

          if (this._annotationEditorUIManager) {
            // Ensure that the Editor buttons, in the toolbar, are updated.
            this.eventBus.dispatch("annotationeditormodechanged", {
              source: this,
              mode: this._annotationEditorMode,
            });
          }

          // In addition to 'disableAutoFetch' being set, also attempt to reduce
          // resource usage when loading *very* long/large documents.
          if (
            pdfDocument.loadingParams.disableAutoFetch ||
            pagesCount > PagesCountLimit.FORCE_LAZY_PAGE_INIT
          ) {
            // XXX: Printing is semi-broken with auto fetch disabled.
            this._pagesCapability.resolve();
            return;
          }
          let getPagesLeft = pagesCount - 1; // The first page was already loaded.

          if (getPagesLeft <= 0) {
            this._pagesCapability.resolve();
            return;
          }
          for (let pageNum = 2; pageNum <= pagesCount; ++pageNum) {
            const promise = pdfDocument.getPage(pageNum).then(
              pdfPage => {
                const pageView = this._pages[pageNum - 1];
                if (!pageView.pdfPage) {
                  pageView.setPdfPage(pdfPage);
                }
                if (--getPagesLeft === 0) {
                  this._pagesCapability.resolve();
                }
              },
              reason => {
                console.error(
                  `Unable to get page ${pageNum} to initialize viewer`,
                  reason
                );
                if (--getPagesLeft === 0) {
                  this._pagesCapability.resolve();
                }
              }
            );

            if (pageNum % PagesCountLimit.PAUSE_EAGER_PAGE_INIT === 0) {
              await promise;
            }
          }
        });

        this.eventBus.dispatch("pagesinit", { source: this });

        pdfDocument.getMetadata().then(({ info }) => {
          if (pdfDocument !== this.pdfDocument) {
            return; // The document was closed while the metadata resolved.
          }
          if (info.Language) {
            this.viewer.lang = info.Language;
          }
        });

        if (this.defaultRenderingQueue) {
          this.update();
        }
      })
      .catch(reason => {
        console.error("Unable to initialize viewer", reason);

        this._pagesCapability.reject(reason);
      });
  }

  renderAnnotation(annotation) {
    if (!this._pages[annotation.pageIndex]) return
    if (this._pages[annotation.pageIndex].compdfAnnotationLayer) {
      this._pages[annotation.pageIndex].compdfAnnotationLayer.renderAnnotation(annotation);
    }
  }

  initBindEvents() {
    const { eventBus } = this;

    eventBus._on("scalechanging", this.webViewerScaleChanging.bind(this));
  }

  // 绑定自定义事件
  bindEvents() {
    const { eventBus } = this;

    eventBus._on("resize", this.webViewerResize.bind(this));
    eventBus._on("pagerendered", this.webViewerPageRendered.bind(this));
    eventBus._on("pagechanging", this.webViewerPageChanging.bind(this));
    eventBus._on("rotationchanging", this.webViewerRotationChanging.bind(this));
    eventBus._on("sidebarviewchanged", this.webViewerSidebarViewChanged.bind(this));
    eventBus._on("presentationmodechanged", this.webViewerPresentationModeChanged.bind(this));
    eventBus._on("updatefindmatchescount", this.webViewerUpdateFindMatchesCount.bind(this));
    eventBus._on("updatefindcontrolstate", this.webViewerUpdateFindControlState.bind(this));

    eventBus._on("annotationsCountChanged", this.webViewerAnnotationsCountChanged.bind(this))
    eventBus._on("distanceChanged", this.webViewerDistanceChanged.bind(this));
    eventBus._on("search", this.webViewerSearch.bind(this));

    eventBus._on("annotationChange", this.handleAnnotationChange.bind(this));
    eventBus._on("createSignature", this.handleCreateSignature.bind(this));
    eventBus._on("handleField", this.handleField.bind(this));
    eventBus._on("pageNumberChanged", this.pageNumberChanged.bind(this));
  }

  bindWindowEvents() {
    const { eventBus, _boundEvents } = this;

    _boundEvents.windowResize = () => {
      eventBus.dispatch("resize", { source: window });
    };
    _boundEvents.windowHashChange = () => {
      eventBus.dispatch("hashchange", {
        source: window,
        hash: document.location.hash.substring(1),
      });
    };
    _boundEvents.windowUpdateFromSandbox = event => {
      eventBus.dispatch("updatefromsandbox", {
        source: window,
        detail: event.detail,
      });
    };

    window.addEventListener("resize", _boundEvents.windowResize);
    window.addEventListener("hashchange", _boundEvents.windowHashChange);
    window.addEventListener(
      "updatefromsandbox",
      _boundEvents.windowUpdateFromSandbox
    );
  }

  async close() {
    if (!this.pdfLoadingTask) {
      return
    }
    const promises = [];

    promises.push(this.pdfLoadingTask.destroy());
    this.pdfLoadingTask = null
    if (this.pdfDocument) {
      this.pdfDocument = null;

      this.pdfThumbnailViewer?.setDocument(null);
      this.pdfViewer.setDocument(null);
      this.pdfLinkService.setViewer(null);
    }
    this.pdfLinkService.externalLinkEnabled = true;
    this.isInitialViewSet = false;
    this.downloadComplete = false;
    this.url = "";
    this.baseUrl = "";
    this._downloadUrl = "";
    this.documentInfo = null;
    this.metadata = null;
    this._contentDispositionFilename = null;
    this._saveInProgress = false;
    this._hasAnnotationEditors = false;

    this.pdfSidebar?.reset();
    this.pdfAttachmentViewer?.reset();

    this.outlines = []
    this.#pwd = ''
    this._pdfId = null
    this.annotations = null

    if (this.findBar && this.findBar.opened) {
      this.findBar.close();
    }

    await Promise.all(promises);
  }

  handleInk(annotation) {
    let rawAnnotation = JSON.parse(JSON.stringify(annotation))
    let inklist = ''
    let inkArray = []
    if (Array.isArray(rawAnnotation.inklist)) {
      inkArray = rawAnnotation.inklist
    }
    else {
      inkArray.push(rawAnnotation.inklist)
    }
    for (let i = 0; i < inkArray.length; i++) {
      inklist = inklist + inkArray[i].replaceAll(';', ',') + '//'
    }
    rawAnnotation.inklist = inklist// rawAnnotation.inklist.replaceAll(';', ',') + '//'
    return rawAnnotation
  }

  async handleAnnotationChange(data) {
    const annotation = data.annotation
    this.annotationHistory.push(annotation)
    const annotateHandles = []
    console.log(annotation)

    if (data.type === 'add') {
      if (Array.isArray(annotation)) {
        const length = annotation.length
        for (let i = 0; i < length; i++) {
          if (Number(annotation[i].pageIndex) + 1 > this.pageCount) continue
          annotation[i].operate = "add-annot"
          await this.handleAnnotations(annotation[i])
          if (this.pdfViewer) {
            this.pdfViewer.renderAnnotation(annotation[i])
          }
          annotation[i].targetPage = annotation[i].pageIndex * 1 + 1
          let currentAnnotation = annotation[i]
          if (currentAnnotation.inklist) {
            currentAnnotation = this.handleInk(currentAnnotation)
          }
          annotateHandles.push(currentAnnotation)
        }
      } else {
        if (Number(annotation.pageIndex) + 1 > this.pageCount) return
        this.handleAnnotations(annotation)

        annotation.targetPage = annotation.pageIndex + 1

        let currentAnnotation = annotation
        if (currentAnnotation.inklist) {
          currentAnnotation = this.handleInk(currentAnnotation)
        }
        annotateHandles.push(currentAnnotation)
      }
    } else {
      const annotations = this.annotations[annotation.pageIndex]
      const index = findIndex(annotation.name, annotations)
      if (data.type === 'delete') {
        annotations.splice(index, 1)
        if (!annotations.length) {
          delete this.annotations[annotation.pageIndex]
        }
        this.messageHandler.sendWithPromise('RemoveAnnot', {
          annotPtr: annotation.annotPtr
        })
      } else {
        annotation.doc = this.#doc
        this.messageHandler.sendWithPromise('EditAnnotation', {
          annotation
        })
        const rawAnnotation = annotations[index]
        annotations[index] = {
          ...rawAnnotation,
          ...annotation
        }
      }
      annotation.targetPage = annotation.pageIndex + 1

      let currentAnnotation = annotation
      if (currentAnnotation.inklist) {
        currentAnnotation = this.handleInk(currentAnnotation)
      }
      annotateHandles.push(currentAnnotation)
    }

    this.eventBus.dispatch('annotationChanged', { annotations: this.annotations })

    // const options = {
    //   method: 'POST',
    //   headers: {
    //     "Content-Type": "application/json",
    //     'Authorization': this._token
    //   },
    //   body: JSON.stringify({
    //     pdfId: this._pdfId,
    //     annotateHandles
    //   }),
    // }
    // fetch(this.optionUrl.webviewBaseUrl + this.optionUrl.editUrl, options)
    //   .then((res) => {
    //     return res.json()
    //   })
    //   .then((data) => {
    //     if (data.code === "200") {
    //       return true
    //     } else {
    //       return false
    //     }
    //   })
    //   .catch((error) => (console.log(error)));
    return true
  }

  handleCreateSignature(data) {
    this.addAnnotations(data)
  }

  webViewerSearch(value) {
    this.findBar?.dispatchEvent("", false, value)
  }

  webViewerDistanceChanged(evt) {
    this.distanceChangedCallback(evt.distance)
  }

  progress(level) {
    if (this.downloadComplete) {
      // Don't accidentally show the loading bar again when the entire file has
      // already been fetched (only an issue when disableAutoFetch is enabled).
      return
    }
    const percent = Math.round(level * 100);
    // When we transition from full request to range requests, it's possible
    // that we discard some of the loaded data. This can cause the loading
    // bar to move backwards. So prevent this by only updating the bar if it
    // increases.
    if (percent <= this.percent) {
      return
    }
    this.percent = percent;
  }

  async flattenPdfDownload() {
    if (this._token && this._pdfId) {
      const data = {
        method: 'POST',
        headers: {
          "Content-Type": "application/json",
          'Authorization': this._token
        },
        body: JSON.stringify({
          pdfId: this._pdfId
        })
      }
      return fetch(this.optionUrl.webviewBaseUrl + this.optionUrl.flattenUrl, data)
        .then(async (res) => {
          const type = res.headers.get("content-type")
          if (type.startsWith("application/json")) {
            const errorObj = await res.json();
            return Promise.reject(errorObj);
          } else {
            return res.blob()
          }
        })
        .then((data) => {
          saveAs(data, this._docName)
          return true
        })
        .catch((error) => {
          console.log(error)
          return false
        });
    } else {
      let saveType = 2
      if (!this.saveAction || this.saveAction === 'set') {
        saveType = 2
      } else if (this.saveAction === 'remove') {
        saveType = 3
      }
      const blobData = await this.messageHandler.sendWithPromise('FlattenPage', {
        doc: this.#doc,
        pagesPtr: this.pagesPtr,
        saveType,
        password: this.#pwd,
        oldPassword: this.#oldPwd
      })
      saveAs(blobData, this._docName)
      return true
    }
  }

  async download(download = true, data, type) {
    if (data) {
      data.forEach(file => {
        saveAs(file.url, file.fileName)
      })
    } else if (this._token && this._pdfId) {
      const data = {
        method: 'POST',
        headers: {
          "Content-Type": "application/json",
          'Authorization': this._token
        },
        body: JSON.stringify({
          pdfId: this._pdfId
        })
      }
      return fetch(this.optionUrl.webviewBaseUrl + this.optionUrl.saveUrl, data)
        .then(async (res) => {
          const type = res.headers.get("content-type")
          if (type.startsWith("application/json")) {
            const errorObj = await res.json();
            return Promise.reject(errorObj);
          } else {
            return res.blob()
          }
        })
        .then((data) => {
          if (download) {
            saveAs(data, this._docName)
          }
          return data
        })
        .catch((error) => {
          console.log(error)
          return false
        });
    } else {
      let saveType = 2
      if (!this.saveAction || this.saveAction === 'set') {
        saveType = 2
      } else if (this.saveAction === 'remove') {
        saveType = 3
      }
      const blobData = await this.messageHandler.sendWithPromise('SaveDocumentByStream', {
        doc: this.#doc,
        saveType: type || saveType,
        password: this.#pwd,
        oldPassword: this.#oldPwd
      })
      if (download) {
        saveAs(blobData, this._docName)
      }
      return blobData
    }
  }

  async exportXfdf(download = true) {
    const xfdf = await this.messageHandler.sendWithPromise('XFDFExportAnnotations', {
      doc: this.#doc
    })
    const blob = new Blob([xfdf], { type: 'application/vnd.adobe.xfdf' })
    if (download) {
      const name = this._docName.replace(/\.pdf$/, '.xfdf')
      saveAs(blob, name)
    }
    return blob
    return fetch(this.optionUrl.webviewBaseUrl + this.optionUrl.downloadAnnotation + "?pdfId=" + this._pdfId, {
      method: 'GET',
      headers: {
        "Content-Type": "application/json",
        'Authorization': this._token
      }
    })
      .then(async (res) => {
        const type = res.headers.get("content-type")
        if (type.startsWith("application/json")) {
          const errorObj = await res.json();
          return Promise.reject(errorObj);
        } else {
          return res.blob()
        }
      })
      .then((data) => {
        const name = this._docName.replace(/\.pdf$/, '.xfdf')
        saveAs(data, name)
        return true
      })
      .catch((error) => {
        console.log(error)
        return false
      });
  }

  setInitialView(
    { scrollMode, spreadMode } = {}
  ) {
    const setRotation = angle => {
      if (isValidRotation(angle)) {
        this.pdfViewer.pagesRotation = angle;
      }
    };
    const setViewerModes = (scroll, spread) => {
      if (isValidScrollMode(scroll)) {
        this.pdfViewer.scrollMode = scroll;
      }
      if (isValidSpreadMode(spread)) {
        this.pdfViewer.spreadMode = spread;
      }
    };
    this.isInitialViewSet = true;

    setViewerModes(scrollMode, spreadMode);

    if (!this.pdfViewer.currentScaleValue) {
      // Scale was not initialized: invalid bookmark or scale was not specified.
      // Setting the default one.
      this.pdfViewer.currentScaleValue = DEFAULT_SCALE_VALUE;
    }
  }

  _scrollUpdate() {
    if (this.pageCount === 0) {
      return;
    }
    this.update();
  }

  update() {
    const visible = this._getVisiblePages();
    const visiblePages = visible.views,
      numVisiblePages = visiblePages.length;

    if (numVisiblePages === 0) {
      return;
    }
    const newCacheSize = Math.max(DEFAULT_CACHE_SIZE, 2 * numVisiblePages + 1);
    this.#buffer.resize(newCacheSize, visible.ids);

    this.pdfRenderingQueue.renderHighestPriority(visible);

    const isSimpleLayout =
      this._spreadMode === SpreadMode.NONE &&
      (this._scrollMode === ScrollMode.PAGE ||
        this._scrollMode === ScrollMode.VERTICAL);
    const currentId = this._currentPageNumber;
    let stillFullyVisible = false;

    for (const page of visiblePages) {
      if (page.percent < 100) {
        break;
      }
      if (page.id === currentId && isSimpleLayout) {
        stillFullyVisible = true;
        break;
      }
    }
    this.#setCurrentPageNumber(
      stillFullyVisible ? currentId : visiblePages[0].id
    );
  }

  _getVisiblePages() {
    const views =
      this._scrollMode === ScrollMode.PAGE
        ? this.#scrollModePageState.pages
        : this._pages,
      horizontal = this._scrollMode === ScrollMode.HORIZONTAL

    return getVisibleElements({
      scrollEl: this.container,
      views,
      sortByVisibility: true,
      horizontal,
    });
  }

  /**
   * @param {number} pageNumber
   */
  isPageVisible(pageNumber) {
    if (!this.pdfDocument) {
      return false;
    }
    if (
      !(
        Number.isInteger(pageNumber) &&
        pageNumber > 0 &&
        pageNumber <= this.pagesCount
      )
    ) {
      console.error(`isPageVisible: "${pageNumber}" is not a valid page.`);
      return false;
    }
    return this._getVisiblePages().ids.has(pageNumber);
  }

  /**
   * @param {number} pageNumber
   */
  isPageCached(pageNumber) {
    if (!this.pdfDocument) {
      return false;
    }
    if (
      !(
        Number.isInteger(pageNumber) &&
        pageNumber > 0 &&
        pageNumber <= this.pagesCount
      )
    ) {
      console.error(`isPageCached: "${pageNumber}" is not a valid page.`);
      return false;
    }
    const pageView = this._pages[pageNumber - 1];
    return this.#buffer.has(pageView);
  }

  requestFullScreenMode() {
    this.pdfPresentationMode?.request();
  }

  cleanup() {
    this._cleanup()
  }

  /**
   * @private
   */
  _cleanup() {
    if (!this.pdfDocument) {
      return; // run cleanup when document is loaded
    }
    for (const pageView of this._pages) {
      if (pageView.renderingState !== RenderingStates.FINISHED) {
        pageView.reset();
      }
    }
    this.pdfThumbnailViewer?.cleanup();
    // We don't want to remove fonts used by active page SVGs.
    this.pdfDocument.cleanup(
      /* keepLoadedFonts = */ this.pdfViewer.renderer === 'svg'
    );
  }

  /**
   * @private
   */
  _cancelRendering() {
    for (const pageView of this._pages) {
      pageView.cancelRendering();
    }
  }

  /**
   * @param {PDFPageView} pageView
   * @returns {Promise<PDFPageProxy | null>}
   */
  async #ensurePdfPageLoaded(pageView) {
    if (pageView.pdfPage) {
      return pageView.pdfPage;
    }
    try {
      const pdfPage = await this.pdfDocument.getPage(pageView.id);
      if (!pageView.pdfPage) {
        pageView.setPdfPage(pdfPage);
      }
      return pdfPage;
    } catch (reason) {
      console.error("Unable to get page for page view", reason);
      return null; // Page error -- there is nothing that can be done.
    }
  }

  #getScrollAhead(visible) {
    if (visible.first?.id === 1) {
      return true;
    } else if (visible.last?.id === this.pagesCount) {
      return false;
    }
    switch (this._scrollMode) {
      case ScrollMode.PAGE:
        return this.#scrollModePageState.scrollDown;
      case ScrollMode.HORIZONTAL:
        return this.scroll.right;
    }
    return this.scroll.down;
  }

  zoomIn(options = null) {
    if (this.pdfViewer.isInPresentationMode) {
      return;
    }
    if (!this.pdfDocument) {
      return;
    }

    options ||= Object.create(null);

    let newScale = this._currentScale;
    if (options.scaleFactor > 1) {
      newScale = Math.min(
        MAX_SCALE,
        Math.round(newScale * options.scaleFactor * 100) / 100
      );
    } else {
      let steps = options.steps ?? 1;
      do {
        newScale = (newScale * DEFAULT_SCALE_DELTA).toFixed(2);
        newScale = Math.ceil(newScale * 10) / 10;
        newScale = Math.min(MAX_SCALE, newScale);
      } while (--steps > 0 && newScale < MAX_SCALE);
    }

    options.noScroll = false;
    this._setScale(newScale, options);
  }

  zoomOut(options = null) {
    if (this.pdfViewer.isInPresentationMode) {
      return;
    }
    if (!this.pdfDocument) {
      return;
    }

    options ||= Object.create(null);

    let newScale = this._currentScale;
    if (options.scaleFactor > 0 && options.scaleFactor < 1) {
      newScale = Math.max(
        MIN_SCALE,
        Math.round(newScale * options.scaleFactor * 100) / 100
      );
    } else {
      let steps = options.steps ?? 1;
      do {
        newScale = (newScale / DEFAULT_SCALE_DELTA).toFixed(2);
        newScale = Math.floor(newScale * 10) / 10;
        newScale = Math.max(MIN_SCALE, newScale);
      } while (--steps > 0 && newScale > MIN_SCALE);
    }

    options.noScroll = false;
    this._setScale(newScale, options);
  }

  refresh(noUpdate = false, updateArgs = Object.create(null)) {
    if (!this.pdfDocument) {
      return;
    }
    console.log(updateArgs)
    for (const pageView of this._pages) {
      pageView.update(updateArgs);
    }
    if (this.#scaleTimeoutId !== null) {
      clearTimeout(this.#scaleTimeoutId);
      this.#scaleTimeoutId = null;
    }
    if (!noUpdate) {
      this.update();
    }
  }

  setFontFile(file) {
    for (const pageView of this._pages) {
      pageView.fontFile = file;
    }
  }

  pageNumberChanged(value) {
    // Note that for `<input type="number">` HTML elements, an empty string will
    // be returned for non-number inputs; hence we simply do nothing in that case.
    if (value !== "") {
      this.pdfLinkService.goToPage(value);
    }

    // Ensure that the page number input displays the correct value, even if the
    // value entered by the user was invalid (e.g. a floating point number).
    if (
      value !== this.pdfViewer.currentPageNumber.toString() &&
      value !== this.pdfViewer.currentPageLabel
    ) {
      this.toolbar && this.toolbar.setPageNumber(
        pdfViewer.currentPageNumber,
        pdfViewer.currentPageLabel
      );
    }
  }

  scrollTo({
    left,
    top
  }) {
    scrollIntoView(this.viewer, {
      left,
      top
    })
  }

  webViewerAnnotationsCountChanged(evt) {
    this.annotationsNumChangedCallback(evt.annotationsCount);
  }

  forceRendering() {
    this.pdfRenderingQueue.printing = !!this.printService;
    this.pdfRenderingQueue.isThumbnailViewEnabled =
      this.pdfSidebar.visibleView === SidebarView.THUMBS;
    this.pdfRenderingQueue.renderHighestPriority();
  }

  forceRenderingPages(currentlyVisiblePages) {
    const visiblePages = currentlyVisiblePages || this._getVisiblePages();
    const scrollAhead = this.#getScrollAhead(visiblePages);
    const preRenderExtra =
      this._spreadMode !== SpreadMode.NONE &&
      this._scrollMode !== ScrollMode.HORIZONTAL;

    const pageView = this.renderingQueue.getHighestPriority(
      visiblePages,
      this._pages,
      scrollAhead,
      preRenderExtra
    );

    if (pageView) {
      this.#ensurePdfPageLoaded(pageView).then(() => {
        this.renderingQueue.renderView(pageView);
      });
      return true;
    }
    return false;
  }

  /**
   * @type {boolean} Whether all pages of the PDF document have identical
   *   widths and heights.
   */
  get hasEqualPageSizes() {
    const firstPageView = this._pages[0];
    for (let i = 1, ii = this._pages.length; i < ii; ++i) {
      const pageView = this._pages[i];
      if (
        pageView.width !== firstPageView.width ||
        pageView.height !== firstPageView.height
      ) {
        return false;
      }
    }
    return true;
  }

  /**
   * @type {number} One of the values in {ScrollMode}.
   */
  get scrollMode() {
    return this._scrollMode;
  }

  /**
   * @param {number} mode - The direction in which the document pages should be
   *   laid out within the scrolling container.
   *   The constants from {ScrollMode} should be used.
   */
  set scrollMode(mode) {
    if (this._scrollMode === mode) {
      return; // The Scroll mode didn't change.
    }
    if (!isValidScrollMode(mode)) {
      throw new Error(`Invalid scroll mode: ${mode}`);
    }
    if (this.pagesCount > PagesCountLimit.FORCE_SCROLL_MODE_PAGE) {
      return; // Disabled for performance reasons.
    }
    this._previousScrollMode = this._scrollMode;

    this._scrollMode = mode;

    this._updateScrollMode(/* pageNumber = */ this._currentPageNumber);
  }

  _updateScrollMode(pageNumber = null) {
    const scrollMode = this._scrollMode,
      viewer = this.viewer;

    viewer.classList.toggle(
      "scrollHorizontal",
      scrollMode === ScrollMode.HORIZONTAL
    )
    viewer.classList.toggle("scrollWrapped", scrollMode === ScrollMode.WRAPPED)

    if (!this.pdfDocument || !pageNumber) {
      return;
    }

    if (scrollMode === ScrollMode.PAGE) {
      this.#ensurePageViewVisible();
    } else if (this._previousScrollMode === ScrollMode.PAGE) {
      // Ensure that the current spreadMode is still applied correctly when
      // the *previous* scrollMode was `ScrollMode.PAGE`.
      this._updateSpreadMode();
    }
    // Non-numeric scale values can be sensitive to the scroll orientation.
    // Call this before re-scrolling to the current page, to ensure that any
    // changes in scale don't move the current page.
    if (this._currentScaleValue && isNaN(this._currentScaleValue)) {
      this._setScale(this._currentScaleValue, { noScroll: true });
    }
    this._setCurrentPageNumber(pageNumber, /* resetCurrentPageView = */ true);
    this.update();
  }

  /**
   * @type {number} One of the values in {SpreadMode}.
   */
  get spreadMode() {
    return this._spreadMode;
  }

  /**
   * @param {number} mode - Group the pages in spreads, starting with odd- or
   *   even-number pages (unless `SpreadMode.NONE` is used).
   *   The constants from {SpreadMode} should be used.
   */
  set spreadMode(mode) {
    if (this._spreadMode === mode) {
      return; // The Spread mode didn't change.
    }
    if (!isValidSpreadMode(mode)) {
      throw new Error(`Invalid spread mode: ${mode}`);
    }
    this._spreadMode = mode;

    this._updateSpreadMode(/* pageNumber = */ this._currentPageNumber);
  }

  _updateSpreadMode(pageNumber = null) {
    if (!this.pdfDocument) {
      return;
    }
    const viewer = this.viewer,
      pages = this._pages;

    if (this._scrollMode === ScrollMode.PAGE) {
      this.#ensurePageViewVisible();
    } else {
      // Temporarily remove all the pages from the DOM.
      viewer.textContent = "";

      if (this._spreadMode === SpreadMode.NONE) {
        for (const pageView of this._pages) {
          viewer.append(pageView.div);
        }
      } else {
        const parity = this._spreadMode - 1;
        let spread = null;
        for (let i = 0, ii = pages.length; i < ii; ++i) {
          if (spread === null) {
            spread = document.createElement("div");
            spread.className = "spread";
            viewer.append(spread);
          } else if (i % 2 === parity) {
            spread = spread.cloneNode(false);
            viewer.append(spread);
          }
          spread.append(pages[i].div);
        }
      }
    }

    if (!pageNumber) {
      return;
    }
    // Non-numeric scale values can be sensitive to the scroll orientation.
    // Call this before re-scrolling to the current page, to ensure that any
    // changes in scale don't move the current page.
    if (this._currentScaleValue && isNaN(this._currentScaleValue)) {
      this._setScale(this._currentScaleValue, { noScroll: true });
    }
    this._setCurrentPageNumber(pageNumber, /* resetCurrentPageView = */ true);
    this.update();
  }

  /**
   * @private
   */
  _getPageAdvance(currentPageNumber, previous = false) {
    switch (this._scrollMode) {
      case ScrollMode.WRAPPED: {
        const { views } = this._getVisiblePages(),
          pageLayout = new Map();

        // Determine the current (visible) page layout.
        for (const { id, y, percent, widthPercent } of views) {
          if (percent === 0 || widthPercent < 100) {
            continue;
          }
          let yArray = pageLayout.get(y);
          if (!yArray) {
            pageLayout.set(y, (yArray ||= []));
          }
          yArray.push(id);
        }
        // Find the row of the current page.
        for (const yArray of pageLayout.values()) {
          const currentIndex = yArray.indexOf(currentPageNumber);
          if (currentIndex === -1) {
            continue;
          }
          const numPages = yArray.length;
          if (numPages === 1) {
            break;
          }
          // Handle documents with varying page sizes.
          if (previous) {
            for (let i = currentIndex - 1, ii = 0; i >= ii; i--) {
              const currentId = yArray[i],
                expectedId = yArray[i + 1] - 1;
              if (currentId < expectedId) {
                return currentPageNumber - expectedId;
              }
            }
          } else {
            for (let i = currentIndex + 1, ii = numPages; i < ii; i++) {
              const currentId = yArray[i],
                expectedId = yArray[i - 1] + 1;
              if (currentId > expectedId) {
                return expectedId - currentPageNumber;
              }
            }
          }
          // The current row is "complete", advance to the previous/next one.
          if (previous) {
            const firstId = yArray[0];
            if (firstId < currentPageNumber) {
              return currentPageNumber - firstId + 1;
            }
          } else {
            const lastId = yArray[numPages - 1];
            if (lastId > currentPageNumber) {
              return lastId - currentPageNumber + 1;
            }
          }
          break;
        }
        break;
      }
      case ScrollMode.HORIZONTAL: {
        break;
      }
      case ScrollMode.PAGE:
      case ScrollMode.VERTICAL: {
        if (this._spreadMode === SpreadMode.NONE) {
          break; // Normal vertical scrolling.
        }
        const parity = this._spreadMode - 1;

        if (previous && currentPageNumber % 2 !== parity) {
          break; // Left-hand side page.
        } else if (!previous && currentPageNumber % 2 === parity) {
          break; // Right-hand side page.
        }
        const { views } = this._getVisiblePages(),
          expectedId = previous ? currentPageNumber - 1 : currentPageNumber + 1;

        for (const { id, percent, widthPercent } of views) {
          if (id !== expectedId) {
            continue;
          }
          if (percent > 0 && widthPercent === 100) {
            return 2;
          }
          break;
        }
        break;
      }
    }
    return 1;
  }

  /**
   * Go to the next page, taking scroll/spread-modes into account.
   * @returns {boolean} Whether navigation occured.
   */
  nextPage() {
    const currentPageNumber = this._currentPageNumber,
      pagesCount = this.pagesCount;

    if (currentPageNumber >= pagesCount) {
      return false;
    }
    const advance =
      this._getPageAdvance(currentPageNumber, /* previous = */ false) || 1;

    this.currentPageNumber = Math.min(currentPageNumber + advance, pagesCount);
    return true;
  }

  /**
   * Go to the previous page, taking scroll/spread-modes into account.
   * @returns {boolean} Whether navigation occured.
   */
  previousPage() {
    const currentPageNumber = this._currentPageNumber;

    if (currentPageNumber <= 1) {
      return false;
    }
    const advance =
      this._getPageAdvance(currentPageNumber, /* previous = */ true) || 1;

    this.currentPageNumber = Math.max(currentPageNumber - advance, 1);
    return true;
  }

  webViewerResize() {
    const { pdfDocument, pdfViewer, pdfRenderingQueue } = this;

    if (pdfRenderingQueue.printing && window.matchMedia("print").matches) {
      // Work-around issue 15324 by ignoring "resize" events during printing.
      return;
    }

    if (!pdfDocument) {
      return;
    }
    const currentScaleValue = pdfViewer.currentScaleValue;
    if (
      currentScaleValue === "auto" ||
      currentScaleValue === "page-fit" ||
      currentScaleValue === "page-width"
    ) {
      // Note: the scale is constant for 'page-actual'.
      pdfViewer.currentScaleValue = currentScaleValue;
    }
    pdfViewer.update();
  }

  webViewerPageRendered({ pageNumber, error }) {
    // If the page is still visible when it has finished rendering,
    // ensure that the page number input loading indicator is hidden.
    /** if (pageNumber === this.page) {
      this.toolbar.updateLoadingIndicatorState(false);
    } */

    // Use the rendered page to set the corresponding thumbnail image.
    if (this.pdfSidebar?.visibleView === SidebarView.THUMBS) {
      const pageView = this.pdfViewer.getPageView(
        /* index = */ pageNumber - 1
      );
      const thumbnailView = this.pdfThumbnailViewer.getThumbnail(
        /* index = */ pageNumber - 1
      );
      if (pageView && thumbnailView) {
        thumbnailView.setImage(pageView);
      }
    }
  }

  get annotationMode() {
    console.log(this._annotationMode)
    return this._annotationMode
  }

  set annotationMode(mode) {
    this._annotationMode = mode;
  }

  switchTool(mode) {
    this.pdfCursorTools.switchTool(mode);
  }

  webViewerScaleChanging(evt) {
    // this.toolbar.setPageScale(evt.presetValue, evt.scale);
    this.pdfViewer.update();
    this.scaleChangedCallback && this.scaleChangedCallback(evt.scale)
  }

  webViewerRotationChanging(evt) {
    this.pdfThumbnailViewer.pagesRotation = evt.pagesRotation;

    this.forceRendering();
    // Ensure that the active page doesn't change during rotation.
    this.pdfViewer.currentPageNumber = evt.pageNumber;
  }

  webViewerPageChanging({ pageNumber }) {
    if (this.pdfSidebar?.visibleView === SidebarView.THUMBS) {
      this.pdfThumbnailViewer.scrollThumbnailIntoView(pageNumber);
    }
    this.pageChangedCallback(this.currentPageNumber)
  }

  webViewerSidebarViewChanged({ view }) {
    this.pdfRenderingQueue.isThumbnailViewEnabled =
      view === SidebarView.THUMBS;
  }

  webViewerPageMode({ mode }) {
    // Handle the 'pagemode' hash parameter, see also `PDFLinkService_setHash`.
    let view;
    switch (mode) {
      case "thumbs":
        view = SidebarView.THUMBS;
        break;
      case "bookmarks":
      case "outline": // non-standard
        view = SidebarView.OUTLINE;
        break;
      case "attachments": // non-standard
        view = SidebarView.ATTACHMENTS;
        break;
      case "layers": // non-standard
        view = SidebarView.LAYERS;
        break;
      case "none":
        view = SidebarView.NONE;
        break;
      default:
        console.error('Invalid "pagemode" hash parameter: ' + mode);
        return;
    }
    this.pdfSidebar.switchView(view, /* forceOpen = */ true);
  }

  async webViewerNamedAction(evt) {
    // Processing a couple of named actions that might be useful, see also
    // `PDFLinkService.executeNamedAction`.
    switch (evt.action) {
      case "Find":
        if (!this.supportsIntegratedFind) {
          if (!evt.close || (evt.close && evt.close === 4 && this.findBar.opened)) {
            this.findBar.toggle();
          }
        }
        break;

      case "FindOpen":
        if (!this.supportsIntegratedFind) {
          if (!evt.close || (evt.close && evt.close === 4 && this.findBar.opened)) {
            this.findBar.open();
          }
        }
        break;

      case "FindClose":
        if (!this.supportsIntegratedFind) {
          if (!evt.close || (evt.close && evt.close === 4 && this.findBar.opened)) {
            this.findBar.close();
          }
        }
        break;

      case "Print":
        const url = await this.download(false, null, 3)
        this.triggerPrinting(url);
        return url
    }
  }

  triggerPrinting(data) {
    let url = data instanceof Blob ? URL.createObjectURL(data) : data
    url = decodeURIComponent(url)
    printJS({
      url,
      printable: url,
      type: 'pdf',
      showModal: false
    });
  }

  webViewerPresentationModeChanged(evt) {
    this.pdfViewer.presentationModeState = evt.state;
  }

  webViewerPageNumberChanged(evt) {
    const pdfViewer = this.pdfViewer;
    // Note that for `<input type="number">` HTML elements, an empty string will
    // be returned for non-number inputs; hence we simply do nothing in that case.
    if (evt.value !== "") {
      this.pdfLinkService.goToPage(evt.value);
    }

    // Ensure that the page number input displays the correct value, even if the
    // value entered by the user was invalid (e.g. a floating point number).
    if (
      evt.value !== pdfViewer.currentPageNumber.toString() &&
      evt.value !== pdfViewer.currentPageLabel
    ) {
      this.toolbar.setPageNumber(
        pdfViewer.currentPageNumber,
        pdfViewer.currentPageLabel
      );
    }
  }
  webViewerScaleChanged(value) {
    this.pdfViewer.currentScaleValue = value;
  }
  webViewerRotateCw() {
    this.rotatePages(90);
  }
  webViewerRotateCcw() {
    this.rotatePages(-90);
  }
  webViewerSwitchScrollMode(mode) {
    this.pdfViewer.scrollMode = mode;
  }
  webViewerSwitchSpreadMode(mode) {
    this.pdfViewer.spreadMode = mode;
  }

  rotatePages(delta) {
    this.pdfViewer.pagesRotation += delta;
    // Note that the thumbnail viewer is updated, and rendering is triggered,
    // in the 'rotationchanging' event handler.
  }

  webViewerUpdateFindMatchesCount({ matchesCount }) {
    if (this.supportsIntegratedFind) {
      this.externalServices.updateFindMatchesCount(matchesCount);
    } else {
      this.findBar.updateResultsCount(matchesCount);
    }
  }

  webViewerUpdateFindControlState({
    state,
    previous,
    matchesCount,
    rawQuery,
  }) {
    if (this.supportsIntegratedFind) {
      this.externalServices.updateFindControlState({
        result: state,
        findPrevious: previous,
        matchesCount,
        rawQuery,
      });
    } else {
      this.findBar.updateUIState(state, previous, matchesCount);
    }
  }

  async setToolMode(mode) {
    this.toolMode = mode;
    // if (mode === 'editor' && !this.fontFileInited) {
    //   await this.initFontFile()
    // }
    this.eventBus.dispatch("toolModeChanged", mode);

    for (let page in this.annotations) {
      for (let i = 0; i < this.annotations[page].length; i++) {
        const annot = this.annotations[page][i]
        const element = document.getElementById(annot['name'])

        if (element) {
          if (this.toolMode !== 'form') {
            if (annot.isHidden == 1) {
              element.style.display = 'none';
              element.style.pointerEvents = 'none';
            } else if (annot.isHidden == 0) {
              element.style.display = 'block';
              element.style.pointerEvents = 'auto';
            }
          } else {
            element.style.display = 'block';
            element.style.pointerEvents = 'auto';
          }
        }
      }
    }
  }

  setPropertyPanel(props) {
    this.eventBus.dispatch("propertyPanelChanged", props);
  }

  setTextProperty(props) {
    this.eventBus.dispatch("textPropertyChanged", props);
  }

  setInitProperty() {
    if (this.activeTool === 'textfield') {
      this.eventBus.dispatch("initTextfieldName");
    } else if (this.activeTool === 'checkbox') {
      this.eventBus.dispatch("initCheckboxName");
    } else if (this.activeTool === 'radiobutton') {
      this.eventBus.dispatch("initRadioButtonName");
    } else if (this.activeTool === 'pushbutton') {
      this.eventBus.dispatch("initPushButtonName");
    }
  }

  handleField(prop) {
    const id = prop.id
    const key = prop.key
    const value = prop.value

    if (!id) return

    let annotationsAll = Object.values(annotationStore.annotationsAll).reduce(function (acc, curr) {
      return acc.concat(curr);
    }, []);
    let targetObj = annotationsAll.find(obj => obj.name === id)
    if (key && (value || value === 0)) {
      let props = {
        name: id
      }
      props[key] = value
      this.eventBus.dispatch("setProperty", props)
    }
    return targetObj
  }

  handleSign(flag, param) {
    if (flag === 'create' && !this.InkSign) {
      this.InkSign = new InkSign({
        pdfViewer: this.pdfViewer,
        eventBus: this.eventBus,
      })
    }
    if (!this.InkSign) return;
    if (flag === 'clear' || flag === 'create') {
      this.InkSign.clear()
    }
    if (flag === 'save') {
      this.InkSign.toImage(param)
    }
    if (flag === 'reset') {
      this.InkSign.reset()
      this.InkSign = null
    }
    if (flag === "update") {
      this.InkSign.update(param)
    }
  }

  handleStamp(data) {
    if (!data) return
    let imgUi = data.stampContainer
    let pageNum = data.page
    let pageView = this.pdfViewer.getPageView(pageNum - 1)
    let rect = null
    let annotData = null

    if (pageView && imgUi) {
      let pageWidth = pageView.viewport.viewBox[2]
      let pageHeight = pageView.viewport.viewBox[3]
      let imgWidth = imgUi.width
      let imgHeight = imgUi.height

      let left = (pageWidth - imgWidth) / 2
      let top = (pageHeight - imgHeight) / 2
      let right = (pageWidth + imgWidth) / 2
      let bottom = (pageHeight + imgHeight) / 2

      rect = {
        left,
        top,
        right,
        bottom,
      }
    }
    if (data.stampType === 'standard') {
      annotData = {
        operate: 'add-annot',
        type: 'stamp',
        pageIndex: pageNum - 1,
        rect,
        stampColor: 'white',
        createDate: new Date(),
        stampType: 'standard',
        content: imgUi.alt
      }
    }
    if (data.stampType === 'dynamic') {
      annotData = {
        operate: 'add-annot',
        type: 'stamp',
        createDate: new Date(),
        pageIndex: pageNum - 1,
        rect,
        content: data.text,
        date: data.time,
        stampColor: data.color,
        stampType: 'text'
      }
    }
    annotData.stampShape = 0
    this.addAnnotations(annotData)
  }

  async compare(data) {
    if (this.webviewerServer) {
      const formData = new FormData()
      for (let key in data) {
        formData.append(key, data[key])
      }
      const options = {
        method: 'POST',
        headers: {
          'Authorization': this._token
        },
        body: formData
      }
      return fetch(this.optionUrl.webviewBaseUrl + this.optionUrl.compareUrl, options)
        .then(async (res) => {
          const type = res.headers.get("content-type")
          if (type.startsWith("application/json")) {
            const errorObj = await res.json();
            return Promise.reject(errorObj);
          } else {
            const filename = res.headers.get("Content-Disposition").split("filename=")[1];
            return { filename, blobData: res.blob(), type };
          }
        })
        .then(async ({ filename, blobData, type }) => {
          if (type === 'application/zip') {
            return await unzipAndReadFiles(blobData)
          } else {
            // return blobData.then(blob => {
            //   const reader = new FileReader();
            //   reader.onload = function(event) {
            //     const content = event.target.result;
            //     console.log(content); // 打印blobData的内容
            //   };
            //   reader.readAsDataURL(blob);
            //   return { filename, url: URL.createObjectURL(blob) };
            // })
            return blobData.then(blob => ({
              filename,
              url: URL.createObjectURL(blob)
            }));
          }
        })
        .then((resp) => resp)
        .catch((error) => ({ error: true, message: error.message }));
    } else {
      const res = await this.messageHandler.sendWithPromise('CompareDocument', {
        data
      })
      if (data.type === 1) {
        return [{
          filename: data.leftFile.name.replace(/(\.pdf)/, `_${this.$t('compare.byComPDFKit')}.pdf`),
          url: URL.createObjectURL(res[0])
        }, {
          filename: data.rightFile.name.replace(/(\.pdf)/, `_${this.$t('compare.byComPDFKit')}.pdf`),
          url: URL.createObjectURL(res[1])
        }, {
          filename: this.$t('compare.byComPDFKit') + '.pdf',
          url: URL.createObjectURL(res[2])
        }
        ]
      } else if (data.type === 2) {
        return {
          filename: this._docName,
          url: URL.createObjectURL(res)
        }
      }
      console.log(res)
      // const { oldData, newData } = res
      // let oldBlobData = new Blob(oldData, { type: "application/pdf" })
      // saveAs(oldBlobData, 'old' + this._docName)


      // let newBlobData = new Blob(newData, { type: "application/pdf" })
      // saveAs(newBlobData, 'new' + this._docName)

    }
  }

  async checkPassword(file) {
    const parameters = {
      cMapUrl: CMAP_URL,
      cMapPacked: true,
      enableXfa: true
    };
    if (typeof file === "string") {
      // URL
      parameters.url = file
    } else if (file && "byteLength" in file) {
      // ArrayBuffer
      parameters.data = file;
    } else if (file.url && file.originalUrl) {
      parameters.url = file.url;
    }
    const loadingTask = getDocument(parameters)
    this.pdfLoadingTask = loadingTask
    this.#pwd = ''
    const getPwd = (pwd) => this.#pwd = pwd
    loadingTask.onPassword = (updateCallback, reason) => {
      this.pdfLinkService.externalLinkEnabled = false;
      this.passwordPrompt.setUpdateCallback(updateCallback, reason, getPwd);
      this.passwordPrompt.open();
    };

    return await loadingTask.promise.then(() => {
      if (this.#pwd) return this.#pwd
      return ''
    }).catch((error) => {
      return false
    })
  }

  handleCreateStatus({
    bool,
    fileData
  }) {
    annotationStore.creating = bool
    annotationStore.fileData = fileData
  }

  async initFontFile() {
    try {
      const response = await fetch('/lib/DroidSansFallbackFull.ttf')
      const blob = await response.blob()
      const file = new File([blob], 'DroidSansFallbackFull.ttf', { type: blob.type })
      this.pdfViewer.setFontFile(file)
      this.fontFileInited = true
    } catch (error) {
      console.log(error)
    }
  }
}

class PDFWorker {
  destroyed = false
  constructor() {
    this._promiseCapability = new PromiseExt()
    this._messageHandler = null
    this._webWorker = null
    this._initialize()
  }

  _initialize() {
    try {
      const workerSrc = "./lib/pdfviewer.worker.js"
      const worker = new Worker(workerSrc)
      const messageHandler = new MessageHandler('main', 'worker', worker)
      this._messageHandler = messageHandler

      const terminateEarly = () => {
        worker.removeEventListener("error", onWorkerError)
        messageHandler.destroy()
        worker.terminate()
        if (this.destroyed) {
          this._promiseCapability.reject(new Error("Worker was destroyed"))
        } else {
          console.error("The worker has been disabled.")
        }
      }

      const onWorkerError = () => {
        if (!this._webWorker) {
          // Worker failed to initialize due to an error. Clean up and fall
          // back to the fake worker.
          terminateEarly();
        }
      }

      worker.addEventListener("error", onWorkerError)

      messageHandler.on("test", (data) => {
        worker.removeEventListener("error", onWorkerError)
        if (this.destroyed) {
          terminateEarly();
          return; // worker was destroyed
        }
        if (data) {
          this._messageHandler = messageHandler
          this._webWorker = worker

          this._promiseCapability.resolve(true);
        }
      })

      messageHandler.on("ready", data => {
        worker.removeEventListener("error", onWorkerError);
        if (this.destroyed) {
          terminateEarly();
          return; // worker was destroyed
        }
        try {
          sendTest();

          messageHandler.send("Ready", true)
        } catch {
          console.error("The worker has been disabled.")
        }
      })

      const sendTest = () => {
        const testObj = new Uint8Array();
        // Ensure that we can use `postMessage` transfers.
        messageHandler.send("test", testObj);
      }

      sendTest()
    } catch (error) {
      console.log(error)
      console.log("The worker has been disabled.")
    }
  }

  get promise() {
    return this._promiseCapability.promise;
  }

  get messageHandler() {
    return this._messageHandler;
  }

  destroy() {
    this.destroyed = true
    if (this._webWorker) {
      // We need to terminate only web worker created resource.
      this._webWorker.terminate()
      this._webWorker = null
    }
    if (this._messageHandler) {
      this._messageHandler.destroy()
      this._messageHandler = null
    }
  }
}

// 解压缩zip文件
function unzipAndReadFiles(zipBlob) {
  return new Promise((resolve, reject) => {
    const jzip = new JSZip();

    jzip.loadAsync(zipBlob)
      .then((zipFiles) => {
        const filePromises = Object.keys(zipFiles.files).map(fileName => {
          return zipFiles.file(fileName).async("uint8array").then(fileData => {
            const blob = new Blob([fileData], { type: 'application/octet-stream' });
            const url = URL.createObjectURL(blob);
            return { fileName, url };
          });
        });

        Promise.all(filePromises)
          .then(resolve)
          .catch(reject);
      })
      .catch((error) => {
        reject(error);
      });
  });
}

export default ComPDFKitViewer
